import click
import numpy as np
import pandas as pd
from scipy.integrate import simps
from scipy.signal import savgol_filter
from scipy.interpolate import interp1d
from numpy import ma as ma

from tqdm import tqdm
from logs import setup_logging
from utils import crossings, get_i_or_default
from analysis.forecast_twl import get_mean_slope, get_intertidal_slope

logger = setup_logging()


def return_first_or_nan(l):
    """
    Returns the first value of a list if empty or returns nan. Used for getting dune/toe and crest values.
    :param l:
    :return:
    """
    if len(l) == 0:
        return np.nan
    else:
        return l[0]


def round_up_to_odd(f):
    """
    https://stackoverflow.com/a/31648815
    :param f:
    :return:
    """
    return int(np.ceil(f) // 2 * 2 + 1)


def volume_change(df_profiles, df_profile_features):
    """
    Calculates how much the volume change there is between prestrom and post storm profiles.
    :param df_profiles:
    :param df_profile_features:
    :return:
    """
    logger.info("Calculating change in beach volume")

    df_vol_changes = pd.DataFrame(
        index=df_profile_features.index.get_level_values("site_id").unique()
    )

    df_profiles = df_profiles.dropna(subset=["z"])
    df_profiles = df_profiles.sort_index()
    sites = df_profiles.groupby(level=["site_id"])

    for site_id, df_site in tqdm(sites):

        params = {}

        for zone in ["dune", "swash"]:
            params[zone] = {}

            for profile_type in ["prestorm", "poststorm"]:
                params[zone][profile_type] = {}

                # Define the edges of the swash and dunes where we want to calculate subaeraial volume.
                if zone == "swash":
                    params[zone][profile_type]["x_min"] = df_profile_features.loc[
                        (site_id, profile_type)
                    ].dune_toe_x
                    params[zone][profile_type]["x_max"] = max(
                        df_profiles.loc[(site_id, profile_type)].index.get_level_values(
                            "x"
                        )
                    )

                    # For profiles with no Dlow value, we take Dhigh as the minimum value to calculate swash
                    if np.isnan(params[zone][profile_type]["x_min"]):
                        params[zone][profile_type]["x_min"] = df_profile_features.loc[
                            (site_id, profile_type)
                        ].dune_crest_x

                elif zone == "dune":
                    params[zone][profile_type]["x_min"] = df_profile_features.loc[
                        (site_id, profile_type)
                    ].dune_crest_x
                    params[zone][profile_type]["x_max"] = df_profile_features.loc[
                        (site_id, profile_type)
                    ].dune_toe_x

                # For profiles with no Dlow value, the dune is undefined and we cannot calculate a dune volume.

                # Calculate subaerial volume based on our x min and maxes
                params[zone][profile_type]["subaerial_vol"] = beach_volume(
                    x=df_profiles.loc[(site_id, profile_type)].index.get_level_values(
                        "x"
                    ),
                    z=df_profiles.loc[(site_id, profile_type)].z.values,
                    x_min=params[zone][profile_type]["x_min"],
                    x_max=params[zone][profile_type]["x_max"],
                )

            # TODO Landward limit still needs to be incorporated

            params[zone]["vol_change"] = (
                params[zone]["poststorm"]["subaerial_vol"]
                - params[zone]["prestorm"]["subaerial_vol"]
            )
            params[zone]["pct_change"] = (
                params[zone]["vol_change"] / params[zone]["prestorm"]["subaerial_vol"]
            )

            df_vol_changes.loc[site_id, "prestorm_{}_vol".format(zone)] = params[zone][
                "prestorm"
            ]["subaerial_vol"]
            df_vol_changes.loc[site_id, "poststorm_{}_vol".format(zone)] = params[zone][
                "poststorm"
            ]["subaerial_vol"]
            df_vol_changes.loc[site_id, "{}_vol_change".format(zone)] = params[zone][
                "vol_change"
            ]
            df_vol_changes.loc[site_id, "{}_pct_change".format(zone)] = params[zone][
                "pct_change"
            ]

    return df_vol_changes


def beach_volume(x, z, x_min=np.NINF, x_max=np.inf):
    """
    Returns the beach volume of a profile, calculated with Simpsons rule
    :param x: x-coordinates of beach profile
    :param z: z-coordinates of beach profile
    :param x_min: Minimum x-coordinate to consider when calculating volume
    :param x_max: Maximum x-coordinate to consider when calculating volume
    :return:
    """
    profile_mask = [True if x_min < x_coord < x_max else False for x_coord in x]
    x_masked = np.array(x)[profile_mask]
    z_masked = np.array(z)[profile_mask]

    if len(x_masked) == 0 or len(z_masked) == 0:
        return np.nan
    else:
        return simps(z_masked, x_masked)


def storm_regime(df_observed_impacts):
    """
    Returns the dataframe with an additional column of storm impacts based on the Storm Impact Scale. Refer to
    Sallenger (2000) for details.
    :param df_observed_impacts:
    :return:
    """
    logger.info("Getting observed storm regimes")

    swash = (df_observed_impacts.dune_pct_change <= 2) & (
        df_observed_impacts.dune_vol_change <= 3
    )
    collision = (df_observed_impacts.dune_pct_change >= 2) | (
        df_observed_impacts.dune_vol_change > 3
    )

    df_observed_impacts.loc[swash, "storm_regime"] = "swash"
    df_observed_impacts.loc[collision, "storm_regime"] = "collision"

    # TODO We may be able to identify observed regimes by looking at the change in crest and toe elevation. This would be useful for
    # locations where we have overwash and cannot calculate the change in volume correctly. Otherwise, maybe it's better to put it in manually.

    return df_observed_impacts


def overwrite_impacts(df_observed_impacts, df_raw_features):
    """
    Overwrites calculated impacts with impacts manually specified in profile_features file
    :param df_raw_profile_features:
    :return:
    """

    # Get manually specified impacts from the profile features ./data/raw/ folder. Note that sites which need to be
    # overwritten with a NaN, use the string 'none' in the csv. This is because when we use the df.update() command,
    # it doesn't overwrite NaN values. So we'll put in the 'none' string, then overwrite that with the NaN.

    df_overwritten_impacts = df_raw_features.rename(
        columns={"observed_storm_regime": "storm_regime"}
    ).storm_regime.to_frame()

    df_observed_impacts.update(df_overwritten_impacts)

    # Replace 'none' with nan
    df_observed_impacts.loc[
        df_observed_impacts.storm_regime == "unknown", "storm_regime"
    ] = np.nan
    return df_observed_impacts


@click.command()
@click.option("--profiles-csv", required=True, help="")
@click.option("--profile-features-crest-toes-csv", required=True, help="")
@click.option("--raw-profile-features-csv", required=True, help="")
@click.option("--output-file", required=True, help="")
def create_observed_impacts(
    profiles_csv, profile_features_crest_toes_csv, raw_profile_features_csv, output_file
):

    logger.info("Creating observed wave impacts")
    logger.info("Importing data")
    df_profiles = pd.read_csv(profiles_csv, index_col=[0, 1, 2])
    df_profile_features = pd.read_csv(profile_features_crest_toes_csv, index_col=[0, 1])

    logger.info("Creating new dataframe for observed impacts")
    df_observed_impacts = pd.DataFrame(
        index=df_profile_features.index.get_level_values("site_id").unique()
    )

    # TODO Review volume change with changing dune toe/crests
    logger.info("Getting pre/post storm volumes")

    df_vol_changes = volume_change(df_profiles, df_profile_features)
    df_observed_impacts = df_observed_impacts.join(df_vol_changes)

    # Classify regime based on volume changes
    df_observed_impacts = storm_regime(df_observed_impacts)

    # Overwrite storm impacts with manually picked impacts
    df_raw_features = pd.read_csv(raw_profile_features_csv, index_col=[0])
    df_observed_impacts = overwrite_impacts(df_observed_impacts, df_raw_features)

    # Calculate change in mean slope
    df_prestorm_mean_slopes = get_mean_slope(
        df_profile_features, df_profiles, profile_type="prestorm"
    )
    df_poststorm_mean_slopes = get_mean_slope(
        df_profile_features, df_profiles, profile_type="poststorm"
    )
    df_diff_mean_slopes = df_poststorm_mean_slopes - df_prestorm_mean_slopes

    # Calculate change in intertidal slope
    df_prestorm_intertidal_slopes = get_intertidal_slope(
        df_profiles, profile_type="prestorm"
    )
    df_poststorm_intertidal_slopes = get_intertidal_slope(
        df_profiles, profile_type="poststorm"
    )
    df_diff_intertidal_slopes = (
        df_poststorm_intertidal_slopes - df_prestorm_intertidal_slopes
    )

    # Rename slope columns and merge into observed impacts
    renames = [
        {"df": df_prestorm_mean_slopes, "new_col_name": "beta_prestorm_mean"},
        {"df": df_poststorm_mean_slopes, "new_col_name": "beta_poststorm_mean"},
        {"df": df_diff_mean_slopes, "new_col_name": "beta_diff_mean"},
        {
            "df": df_prestorm_intertidal_slopes,
            "new_col_name": "beta_prestorm_intertidal",
        },
        {
            "df": df_poststorm_intertidal_slopes,
            "new_col_name": "beta_poststorm_intertidal",
        },
        {"df": df_diff_intertidal_slopes, "new_col_name": "beta_diff_intertidal"},
    ]

    for rename in renames:
        rename["df"].rename(
            {"beta": rename["new_col_name"]}, axis="columns", inplace=True
        )

    # Join all our slopes into the observed impacts
    df_observed_impacts = pd.concat(
        [
            df_prestorm_mean_slopes,
            df_poststorm_mean_slopes,
            df_diff_mean_slopes,
            df_prestorm_intertidal_slopes,
            df_poststorm_intertidal_slopes,
            df_diff_intertidal_slopes,
            df_observed_impacts,
        ],
        axis=1,
    )

    # Calculate change in beach width
    df_width_msl_prestorm = get_beach_width(
        df_profile_features,
        df_profiles,
        profile_type="prestorm",
        ele=0,
        col_name="width_msl_prestorm",
    )
    df_width_msl_poststorm = get_beach_width(
        df_profile_features,
        df_profiles,
        profile_type="poststorm",
        ele=0,
        col_name="width_msl_poststorm",
    )
    df_width_msl_change_m = (df_width_msl_poststorm - df_width_msl_prestorm).rename(
        "width_msl_change_m"
    )
    df_width_msl_change_pct = (
        df_width_msl_change_m / df_width_msl_prestorm * 100
    ).rename("width_msl_change_pct")

    # Join beach width change onto observed impacts dataframe
    df_observed_impacts = pd.concat(
        [
            df_observed_impacts,
            df_width_msl_prestorm,
            df_width_msl_poststorm,
            df_width_msl_change_m,
            df_width_msl_change_pct,
        ],
        axis=1,
    )

    # Save dataframe to csv
    df_observed_impacts.to_csv(output_file, float_format="%.4f")

    logger.info("Saved to %s", output_file)
    logger.info("Done!")


def get_beach_width(df_profile_features, df_profiles, profile_type, ele, col_name):
    df_x_position = (
        df_profiles.xs(profile_type, level="profile_type")
        .dropna(subset=["z"])
        .groupby("site_id")
        .apply(
            lambda x: get_i_or_default(
                crossings(
                    profile_x=x.index.get_level_values("x").tolist(),
                    profile_z=x.z.tolist(),
                    constant_z=ele,
                ),
                -1,
                default=np.nan,
            )
        )
        .rename("x_position")
    )
    df_x_prestorm_dune_toe = df_profile_features.xs(
        "prestorm", level="profile_type"
    ).dune_toe_x
    df_width = (df_x_position - df_x_prestorm_dune_toe).rename(col_name)
    return df_width
